#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
This module provides the CKIP Transformers NLP drivers.
"""

__author__ = "Mu Yang <http://muyang.pro>"
__copyright__ = "2021 CKIP Lab"
__license__ = "GPL-3.0"

from .driver import (
    CkipWordSegmenter,
    CkipPosTagger,
    CkipNerChunker,
)
