#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
The CKIP Transformers.
"""

__author_name__ = "Mu Yang"
__author_email__ = "emfomy@gmail.com"
__copyright__ = "2021 CKIP Lab"

__title__ = "CKIP Transformers"
__version__ = "0.3.0"
__description__ = "CKIP Transformers"
__license__ = "GPL-3.0"

__url__ = "https://ckip-transformers.readthedocs.io"
__download_url__ = __url__ + "/tarball/" + __version__
